#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <jabber/jabber.h>
#include <etherx.h>
#include "genhash.h"

#define NSCHECK(x,n) (j_strcmp(xmlnode_get_attrib(x,"xmlns"),n) == 0)

typedef struct terror_struct
{
    int code;
    char msg[64];
} terror;

#define TERROR_BAD (terror){400,"Bad Request"}
#define TERROR_AUTH (terror){401,"Unauthorized"}
#define TERROR_PAY (terror){402,"Payment Required"}
#define TERROR_FORBIDDEN (terror){403,"Forbidden"}
#define TERROR_NOTFOUND (terror){404,"Not Found"}
#define TERROR_NOTALLOWED (terror){405,"Not Allowed"}
#define TERROR_NOTACCEPTABLE (terror){406,"Not Acceptable"}
#define TERROR_REGISTER (terror){407,"Registration Required"}
#define TERROR_REQTIMEOUT (terror){408,"Request Timeout"}
#define TERROR_CONFLICT (terror){409,"Conflict"}

#define TERROR_INTERNAL (terror){500,"Internal Server Error"}
#define TERROR_NOTIMPL (terror){501,"Not Implemented"}
#define TERROR_EXTERNAL (terror){502,"Remote Server Error"}
#define TERROR_UNAVAIL (terror){503,"Service Unavailable"}
#define TERROR_EXTTIMEOUT (terror){504,"Remote Server Timeout"}


#define NS_CLIENT "jabber:client"
#define NS_SERVER "jabber:server"
#define NS_AUTH "jabber:iq:auth"
#define NS_REGISTER "jabber:iq:register"
#define NS_ROSTER "jabber:iq:roster"
#define NS_OFFLINE "jabber:x:offline"
#define NS_AGENT "jabber:iq:agent"
#define NS_AGENTS "jabber:iq:agents"
#define NS_DELAY "jabber:x:delay"
#define NS_VERSION "jabber:iq:version"
#define NS_TIME "jabber:iq:time"
#define NS_VCARD "vcard-temp"

#define MAPI_VARAUTH -10
#define MAPI_VARREGISTER -11

typedef enum {C_SET, C_GET, C_DEL, C_CHECK, C_INIT} command;

typedef enum {S_BOUNCE, S_IGNORE, S_AUTHED} sreturn;

typedef enum {P_SESSION, P_UNKNOWN, P_OFFLINE, P_SERVER, P_DELIVER, P_SHUTDOWN, PS_IN, PS_OUT, PS_END} mphase;
typedef enum {M_PASS, M_IGNORE, M_HANDLED} mreturn;


typedef struct udata_struct *udata, _udata;
typedef struct session_struct *session, _session;

typedef struct mapi_struct
{
    jpacket packet;
    mphase phase;
    udata user;
    session s;
    int variant;
} *mapi, _mapi;

typedef mreturn (*mcall)(mapi m, void *arg);

typedef struct mlist_struct
{
    mcall c;
    void *arg;
    unsigned char mask;
    struct mlist_struct *next;
} *mlist, _mlist;

/* contains a list of module function pointers */
typedef struct mapi_master
{
    mphase p;
    mlist l;
    struct mapi_master *next;
} *mmaster, _mmaster;

mmaster js_mapi_master(mphase p);

typedef struct xdb_struct
{
    char *ns;
    xmlnode data;
    int cache;
    struct xdb_struct *next;
} *xdb, _xdb;

typedef int (*xcall)(int set, xdb x, udata u, void *arg);

/* contains a list of xdb handler function pointers */
typedef struct xlist_struct
{
    xcall c;
    void *arg;
    struct xlist_struct *next;
} *xlist, _xlist;

void js_xdb_register(xcall c, void *arg);
xmlnode js_xdb_get(udata user, char *ns);
void js_xdb_set(udata user, char *ns, xmlnode x);

void js_listen(xmlnode x);

ehandler js_names(int cmd, char *name);


struct udata_struct
{
    char *user;
    session sessions;
    int scount, ref;
    xdb x_cache;
    ppdb p_cache;
    pool p;
    struct udata_struct *next;
};

xmlnode js_config(char *query);

udata js_user(char *user);
void js_users_exit(void);
void js_deliver(jpacket p);


typedef void (*session_onSend)(session s, jpacket p, void *arg);

struct session_struct
{
    /* general session data */
    char *res;
    jid id, uid;
    udata u;
    xmlnode presence;
    int priority, roster;
    int c_in, c_out;

    /* mechanics */
    pool p;
    pth_msgport_t mp_from, mp_to;
    int exit_flag;
    mlist m_in, m_out, m_end;
    thread t;

    /* send handler */
    session_onSend send;
    void *arg;

    struct session_struct *next;
};

session js_session_new(jid owner, session_onSend send, void *arg);
void *js_session_main(void *arg);
void js_session_end(session s, char *reason);
session js_session_get(udata user, char *res);
session js_session_primary(udata user);

void js_conn_connect(thread t, int sock, struct sockaddr_in sa);
void js_conn_negotiate(thread t, xmlnode in, int type);
xmlstream_onNode js_conn_namespace(command cmd, char *ns, xmlstream_onNode handler);

void *js_server_main(void *arg);
void *js_offline_main(void *arg);
void *js_unknown_main(void *arg);
void *js_users_main(void *arg);
void *js_debug_main(void *arg);

typedef struct {
    pth_message_t head; /* the standard pth message header */
    jpacket p;
} _jpq, *jpq;

void js_psend(pth_msgport_t mp, jpacket p); /* sends p to a pth message port */

xmlnode js_pres_new(int type, char *to, char *status); /* creates a presence template */
xmlnode js_iq_new(int type, char *ns); /* creates an iq template */
int js_priority(xmlnode x);
xmlnode js_iqreply(xmlnode x);
void js_error(xmlnode x, terror terr); /* modifies x for the error */
void js_bounce(xmlnode x, terror terr); /* logic to bounce packets w/o looping, eats x and delivers error */
char *js_timestamp(void); /* returns current timestamp */
void js_delay(xmlnode msg, char *reason);

int js_config_name(command cmd, char *name);
int js_config_load(char *file); /* load the config file */
