#include "jserver.h"
#include "dlfcn.h"

/* loads the shared object and calls the init function */
void mod_dso_load(char *so, char *init)
{
    void (*init_h)(void);
    void *so_h;
    char *dlerr;

    if(so == NULL || init == NULL) return;

    log_debug(ZONE,"mod_dso_load %s",so);

    so_h = dlopen(so,RTLD_LAZY);
    dlerr = dlerror();
    if(dlerr != NULL)
    {
        log_error("jserver","mod_dso loading failed: %s",so,dlerr);
        return;
    }

    init_h = dlsym(so_h,init);
    dlerr = dlerror();
    if(dlerr != NULL)
    {
        log_error("jserver","mod_dso calling %s in %s failed: %s",init,so,dlerr);
        dlclose(so_h);
        return;
    }
    (init_h)();
}

/* loads the dso components listed in the config */
void mod_dso(void)
{
    void (*init_h)(void);
    void *so_h;
    char *dlerr;
    xmlnode cfg;

    log_debug(ZONE,"mod_dso");

    cfg = js_config("dso");
    if(cfg == NULL)
        return;

    for(cfg = xmlnode_get_firstchild(cfg); cfg != NULL; cfg = xmlnode_get_nextsibling(cfg))
        if(xmlnode_get_type(cfg) == NTYPE_TAG)
            mod_dso_load(xmlnode_get_data(cfg), xmlnode_get_name(cfg));
}
