#include "jserver.h"

/* a master list which contains the module lists for each phase */
mmaster js_mapi_master(mphase p)
{
    static mmaster master = NULL;
    mmaster cur;

    log_debug(ZONE,"mapi_master %d",p);

    for(cur = master; cur != NULL; cur = cur->next)
        if(cur->p == p) break;

    if(cur == NULL)
    {
        cur = malloc(sizeof(_mmaster));
        cur->p = p;
        cur->l = NULL;
        cur->next = master;
        master = cur;
    }

    return cur;
}

/* let a module register a new call for this phase */
void js_mapi_register(mphase p, mcall c, void *arg)
{
    mlist newl, curl;
    mmaster master;

    if(c == NULL) return;

    log_debug(ZONE,"mapi_register %d",p);

    newl = malloc(sizeof(_mlist));
    newl->c = c;
    newl->arg = arg;
    newl->mask = 0x00;
    newl->next = NULL;

    /* save the new list item in the master list for this phase */
    master = js_mapi_master(p);
    if(master->l == NULL)
    {
        master->l = newl;
    }else{
        /* append to end of call list */
        for(curl = master->l; curl->next != NULL; curl = curl->next);
        curl->next = newl;
    }
}

/* let a module register a new call for this session phase */
void js_mapi_session(mphase p, session s, mcall c, void *arg)
{
    mlist newl, curl, *curs;

    if(c == NULL || s == NULL) return;

    log_debug(ZONE,"mapi_register_session %d",p);

    newl = pmalloc(s->p, sizeof(_mlist));
    newl->c = c;
    newl->arg = arg;
    newl->mask = 0x00;
    newl->next = NULL;

    /* save the new list item in the master list for this phase */
    switch(p)
    {
        case PS_IN:
            curs = &(s->m_in);
            break;
        case PS_OUT:
            curs = &(s->m_out);
            break;
        case PS_END:
            curs = &(s->m_end);
            break;
        default:
            /* dork */
            return;
    }

    if(*curs == NULL)
    {
        *curs = newl;
    }else{
        /* append to end of call list */
        for(curl = *curs; curl->next != NULL; curl = curl->next);
        curl->next = newl;
    }
}

int js_mapi_call(mphase phase, mlist l, jpacket packet, udata user, session s, int variant)
{
    _mapi m;
    mreturn ret;

    if(l == NULL) return 0;

    log_debug(ZONE,"mapi_call %d",phase);

    m.phase = phase;
    m.packet = packet;
    m.user = user;
    m.s = s;
    m.variant = variant;

    for(;l != NULL; l = l->next)
    {
        if(packet != NULL && (packet->type & l->mask) == packet->type) continue;
        switch((*(l->c))(&m, l->arg))
        {
            case M_IGNORE:
                /* this module is ignoring this packet->type */
                l->mask |= packet->type;
                break;
            case M_HANDLED:
                return 1;
        }
    }

    return 0;
}

