#include "jserver.h"

thread mod_log__session = NULL;
thread mod_log__packet = NULL;

void mod_log_init()
{
    int fd;
    char *file;

    file = xmlnode_get_data(js_config("log/session"));
    fd = open(file, O_WRONLY | O_CREAT | O_APPEND, 0600);
    if(file == NULL || fd < 0)
    {
        log_error("jserver","unable to open session log file");
        return;
    }
    mod_log__session = tstream_new(fd,"sessionlogger",NULL,NULL,NULL);

    file = xmlnode_get_data(js_config("log/packet"));
    fd = open(file, O_WRONLY | O_CREAT | O_APPEND, 0600);
    if(file == NULL || fd < 0)
    {
        log_error("jserver","unable to open packet log file");
        return;
    }
    mod_log__packet = tstream_new(fd,"packetlogger",NULL,NULL,NULL);
}

/* logs session characteristics */
mreturn mod_log_session_end(mapi m, void *arg)
{
    pth_attr_t attr;
    pth_time_t started, total;
    static char log[1024];
    int size;
    time_t t;

    if(mod_log__session == NULL)
        return;

    log_debug("mod_log","creating session log entry");

    attr = pth_attr_of(m->s->t->id);
    pth_attr_get(attr,PTH_ATTR_TIME_SPAWN,&started);
    pth_attr_get(attr,PTH_ATTR_TIME_RAN,&total);
    pth_attr_destroy(attr);
    t = time(NULL);

    /* TIMESTAMP TOTALSECONDS PACKETSFROM PACKETSTO USER@HOST/RES */
    size = snprintf(log, 1024, "%s %d %d %d %s\n",
        create_log_timestamp(),
        t - started.tv_sec,
        m->s->c_in,
        m->s->c_out,
        jid_full(m->s->id));

    tstream_write(mod_log__session,log,size);

    log_debug("CPU","%s used %d.%d CPU time",jid_full(m->s->id),total.tv_sec,total.tv_usec);

    return M_PASS;
}

char *_mod_log_ptype(int type)
{
    static char ret[10];

    ret[0] = '\0';
    switch(type)
    {
        case JPACKET_MESSAGE:
            strcat(ret,"message");
            break;
        case JPACKET_PRESENCE:
            strcat(ret,"presence");
            break;
        case JPACKET_IQ:
            strcat(ret,"iq");
            break;
        case JPACKET_S10N:
            strcat(ret,"s10n");
            break;
    }

    return (char *)ret;
}

char *_mod_log_null(char *data)
{
    static char ret[] = "-";

    if(data == NULL)
        return (char *)ret;
    else
        return data;
}

/* logs packets */
mreturn mod_log_packet(mapi m, void *arg)
{
    static char log[1024];
    int size;

    if(m->packet == NULL || mod_log__packet == NULL)
        return;

    log_debug("mod_log","creating packet log entry");

    /* PACKET TYPE FROM TO */
    size = snprintf(log, 1024, "%s %s %s %s %s\n",
        create_log_timestamp(),
        _mod_log_ptype(m->packet->type),
        _mod_log_null(xmlnode_get_attrib(m->packet->x,"type")),
        _mod_log_null(jid_full(m->packet->from)),
        _mod_log_null(jid_full(m->packet->to)));

    tstream_write(mod_log__packet,log,size);

    return M_PASS;
}

mreturn mod_log_session(mapi m, void *arg)
{
    js_mapi_session(PS_END, m->s, mod_log_session_end, NULL);

    return M_PASS;
}

void mod_log(void)
{
    log_debug("mod_log","init");
    mod_log_init();
    if(mod_log__session != NULL)
        js_mapi_register(P_SESSION, mod_log_session, NULL);
    if(mod_log__packet != NULL)
        js_mapi_register(P_DELIVER, mod_log_packet, NULL);
}

