#include "jserver.h"

mreturn mod_vcard_set(mapi m, void *arg)
{
    xmlnode vcard;

    if(m->packet->type != JPACKET_IQ) return M_IGNORE;
    if(m->packet->to != NULL || !NSCHECK(m->packet->iq,NS_VCARD)) return M_PASS;

    vcard = js_xdb_get(m->user, NS_VCARD);

    switch(jpacket_subtype(m->packet))
    {
        case JPACKET__GET:
            log_debug("mod_vcard","handling get request");
            xmlnode_put_attrib(m->packet->x,"type","result");

            /* insert the vcard into the result */
            xmlnode_insert_node(m->packet->iq, xmlnode_get_firstchild(vcard));
            jpacket_reset(m->packet);

            /* send to the user */
            js_psend(m->s->mp_to,m->packet);

            break;
        case JPACKET__SET:
            log_debug("mod_vcard","handling set request");

            /* save the changes */
            log_debug(ZONE,"VCARD: %s",xmlnode2str(m->packet->iq));
            js_xdb_set(m->user,NS_VCARD,xmlnode_dup(m->packet->iq));

            /* send to the user */
            js_iqreply(m->packet->x);
            jpacket_reset(m->packet);
            js_psend(m->s->mp_to,m->packet);

            break;
        default:
            xmlnode_free(m->packet->x);
            break;
    }
    return M_HANDLED;
}

mreturn mod_vcard_reply(mapi m, void *arg)
{
    xmlnode vcard, item;
    char *sub;

    if(m->packet->type != JPACKET_IQ) return M_IGNORE;
    if(!NSCHECK(m->packet->iq,NS_VCARD)) return M_PASS;

    /* first, is this a valid request? */
    switch(jpacket_subtype(m->packet))
    {
        case JPACKET__RESULT:
        case JPACKET__ERROR:
            return M_PASS;
        case JPACKET__SET:
            js_error(m->packet->x,TERROR_NOTALLOWED);
            return M_HANDLED;
    }

    log_debug("mod_vcard","handling query for user %s",m->user->user);

    /* get this guys vcard info */
    vcard = js_xdb_get(m->user, NS_VCARD);

    /* check permissions, XXX hack for now till vcard draft 
    if(xmlnode_get_tag(vcard,"public") == NULL)
    { /* they must be in our roster to get our vcard 
        item = js_xdb_get(m->user, NS_ROSTER);
        item = jid_nodescan(m->packet->from,item);
        sub = xmlnode_get_attrib(item,"subscription");
        if(sub == NULL || j_strcmp(sub,"none") == 0 || j_strcmp(sub,"to") == 0)
        {
            js_error(m->packet->x,TERROR_FORBIDDEN);
            return;
        }
    }*/

    js_iqreply(m->packet->x);
    jpacket_reset(m->packet);
    xmlnode_insert_tag_node(m->packet->x,vcard);
    js_deliver(m->packet);

    return M_HANDLED;
}

mreturn mod_vcard_session(mapi m, void *arg)
{
    js_mapi_session(PS_OUT,m->s,mod_vcard_set,NULL);
    js_mapi_session(PS_IN,m->s,mod_vcard_reply,NULL);
    return M_PASS;
}

void mod_vcard(void)
{
    js_mapi_register(P_SESSION,mod_vcard_session,NULL);
    js_mapi_register(P_OFFLINE,mod_vcard_reply,NULL);
}


