#include "jserver.h"
#include "svc_client_ssl.h"

typedef struct svc_client_data_struct
{
    session s;
    char *server;
    char *id;
} *clidata, _clidata;

void svc_client_onSend(session s, jpacket p, void *arg)
{
    thread t_socket = (thread)arg;
    tstream ts;
    clidata cd;

    if(p == NULL)
    { /* signal to exit from the session */
        if(arg != NULL)
        {
            log_debug(ZONE,"session exited externally");
            ts = (tstream)(t_socket->data);
            cd = (clidata)(ts->data_aux1);
            cd->s = NULL; /* so that the cleanup handler knows the session is already gone, preventing race condition */
            xmlstream_error(t_socket,"forced disconnect");
        }
        return;
    }

    if(p->from != NULL)
        xmlnode_put_attrib(p->x, "from", jid_full(p->from));

    /* hide the to/from if it's this resource */
    if(jid_cmp(p->to,s->id) == 0)
        xmlnode_hide_attrib(p->x, "to");
    if(p->type != JPACKET_MESSAGE && jid_cmp(p->from,s->id) == 0)
        xmlnode_hide_attrib(p->x, "from");

    log_debug(ZONE,"SVC_BASIC: got a packet[%s] to send to the user[%d]",xmlnode2str(p->x),t_socket);

    xmlstream_write(t_socket, p->x);
}

void svc_client_cleanup(void *arg)
{
    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);
    clidata cd = (clidata)(ts->data_aux1);
#ifdef HAVE_OPENSSL
	SSL *ssl = (SSL *)(ts->data_aux4);

	if(ssl != NULL)
	{
		SSL_set_shutdown(ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
	    SSL_free(ssl);
	    ERR_remove_state(0);
	}
#endif
    log_debug(ZONE,"SVC_BASIC: cleanup %X",cd);
    if(cd->s != NULL)
    {
        cd->s->arg = NULL; /* so that onSend knows the thread is gone, preventing race condition */
        js_session_end(cd->s,"disconnected");
    }
}

/* the connection/stream has passed all the test by the server when it get's here */
void svc_client_onNode(thread t, xmlnode in, int type)
{
    tstream ts = (tstream)(t->data);
    clidata cd = (clidata)(ts->data_aux1);
    xmlnode query;
    jpacket p;
    jid id;
    xmlnode x;
    tstream_buffer b;
    sreturn sr;

    log_debug(ZONE,"SVC_BASIC: socket %d thread %X onNode %X",ts->fd,t->id,cd);

    if(type == XMLSTREAM_ROOT)
    {
        pth_cleanup_push(svc_client_cleanup,(void *)t);
        cd = ts->data_aux1 = pmalloc(t->p, sizeof(_clidata));
        cd->s = NULL;
        cd->server = pstrdup(t->p,xmlnode_get_attrib(in, "to"));
        x = xmlstream_header(NS_CLIENT,NULL,xmlnode_get_attrib(in, "to"));
        cd->id = pstrdup(t->p,xmlnode_get_attrib(x,"id"));
        xmlstream_write_header(t, x);
        return;
    }

    if(type == XMLSTREAM_CLOSE)
    {
        /* the connection was closed */
        return;
    }

    if(type == XMLSTREAM_BOUNCE)
    {
        /* XXX leftovers that didn't get sent */
        log_debug(ZONE,"leftovers: %s",xmlnode2str(in));
        return;
    }

    if(type == XMLSTREAM_ERROR)
    {
        /* the connection was closed */
        xmlstream_error(t, "Invalid XML Received");
        return;
    }

    if(cd->s == NULL)
    {
        xmlnode_put_attrib(in,"sid",cd->id); /* session ID used by auth stuff */
        p = jpacket_new(in);
        query = p->iq; /* so we can access the query data later */
        sr = js_service_prescreen(p);
        xmlnode_hide_attrib(in,"sid");
        switch(sr)
        {
            case S_BOUNCE:
                log_debug(ZONE,"prescreening returned data, bouncing back");
                xmlstream_write(t,xmlnode_dup(in));
                break;
            case S_AUTHED:
                xmlstream_write(t,xmlnode_dup(in));
                /* make a temp jid for this new user */
                id = jid_new(xmlnode_pool(in),cd->server);
                jid_set(id,xmlnode_get_tag_data(query,"username"),JID_USER);
                jid_set(id,xmlnode_get_tag_data(query,"resource"),JID_RESOURCE);
                log_debug(ZONE,"creating new session for %s",jid_full(id));
                cd->s = js_session_new(id, svc_client_onSend, (void *)t);
                break;
        }
        return;
    }

    log_debug(ZONE,"sending packet to session: %s",xmlnode2str(in));
    js_psend(cd->s->mp_from, jpacket_new(xmlnode_dup(in)));
}

void svc_client(void)
{
    log_debug(ZONE,"svc_client: init()");

    js_conn_namespace(C_SET,NS_CLIENT,(xmlstream_onNode)svc_client_onNode);
#ifdef HAVE_OPENSSL
	log_debug(ZONE, "svc_client: init() SSL functions");
	tlisten_new(5223, NULL, svc_ssl_connect, NULL);
#endif
}


