#include "svc_http.h"

/* XXX fix for incoming data across multiple calls/reads, store unparsed buffer
 * 
 * 
 */

/* this is where we send the request on to the right handler */
void svc_http_breakdance(hrq rq)
{
    if(strncmp(rq->path,"/presence",9) == 0) svc_http_presence(rq);
    else if(strncmp(rq->path,"/message",8) == 0) svc_http_message(rq);
    else if(strncmp(rq->path,"/login",8) == 0) svc_http_login(rq);
    else if(strncmp(rq->path,"/auth",8) == 0) svc_http_login(rq);
    else if(strncmp(rq->path,"/roster",8) == 0) svc_http_roster(rq);
    else if(strncmp(rq->path,"/XML",8) == 0) svc_http_xml(rq);
    else if(strncmp(rq->path,"/RPC2",8) == 0) svc_http_rpc(rq);
    else svc_http_error(rq,404);
}

void svc_http_parse(hrq rq)
{
    char *qmark, *file, *dot;

    if(j_strcmp(xmlnode_get_name(rq->x),"POST") == 0)
        svc_http_qs2xml(xmlnode_insert_tag(rq->x,"FORM"),xmlnode_get_data(rq->x));

    /* trim the :port from the host header */
    file = xmlnode_get_attrib(rq->x,"Host");
    if(file != NULL && (dot = strstr(file,":")) != NULL)
        *dot = '\0';

    /* parse any query string */
    qmark = strstr(rq->path,"?");
    if(qmark != NULL)
    {
        *qmark = '\0';
        qmark++;
        rq->q = xmlnode_new_tag_pool(rq->p,"q");
        svc_http_qs2xml(rq->q,qmark);
    }

    /* make the path friendlier */
    svc_http_strdecode(rq->path);
    file = strrchr(rq->path,'/');
    if(file != NULL && rq->file != rq->path)
    { /* file must be after a second slash */
        file++;
        rq->file = pstrdup(rq->p,file);
        dot = strrchr(rq->file,'.');
        if(dot != NULL)
        {
            *dot = '\0';
            dot++;
            rq->ext = dot;
        }
    }else{
        dot = strrchr(rq->path,'.');
        if(dot != NULL)
        {
            dot++;
            rq->ext = dot;
        }
    }

    log_debug(ZONE,"received request [%s] with query [%s] to %s",xmlnode2str(rq->x),xmlnode2str(rq->q),rq->path);
}

/* when the connection dies, clean up the mess */
void svc_http_cleanup(void *arg)
{
    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);
    hrq rq = (hrq)(ts->data);
 
    log_debug(ZONE,"svc_http_cleanup");
    free(rq->buffer);
}

/* read in the http request */
void svc_http_read(thread t, void *buffer, int bufsz)
{
    char *cbuffer, *cur, *uri, *proto, *header;
    int i, cl;
    tstream ts = (tstream)(t->data);
    hrq rq = (hrq)(ts->data);

    if(rq->phase == PHASE_INIT)
    {
        pth_cleanup_push(svc_http_cleanup, (void *)t);
        rq->phase = PHASE_HEAD;
    }

    if(bufsz <= 0) return;

    if(rq->buffer != NULL)
    { /* old data yet, combine and use that */
        cur = malloc(strlen(rq->buffer) + bufsz + 1);
        *cur = '\0';
        strcat(cur,rq->buffer);
        strcat(cur,buffer);
        free(rq->buffer);
        rq->buffer = buffer = cur;
        bufsz = strlen(buffer);
    }

    log_debug(ZONE,"reading data [%s]",buffer);

    /* read http header */
    cur = cbuffer = (char *)buffer;
    for(i = 0; rq->phase == PHASE_HEAD && (i < bufsz); i++)
    {
        if(cbuffer[i] == '\r' || cbuffer[i] == '\n')
        {
            cbuffer[i] = '\0';
            if(strlen(cur) > 0)
            {
                rq->rn = 0;

                if(rq->x == NULL)
                { /* this is the first line, GET /foo HTTP/x.x */
                    uri = strstr(cur," ");
                    *uri = '\0';
                    ++uri;
                    proto = strstr(uri," ");
                    *proto = '\0';
                    ++proto;
                    rq->x = xmlnode_new_tag_pool(rq->p,cur);
                    rq->path = pstrdup(rq->p,uri);
                }else{
                    header = strstr(cur,": ");
                    if(header != NULL)
                    {
                        *header = '\0';
                        ++header;
                        *header = '\0';
                        ++header;
                    }
                    xmlnode_put_attrib(rq->x,cur,header);
                }
            }else{
                rq->rn++;
                if(rq->rn == 3)
                    if(j_strcmp(xmlnode_get_name(rq->x),"POST") == 0 && xmlnode_get_attrib(rq->x,"Content-length") != NULL)
                    {
                        rq->clen = atoi(xmlnode_get_attrib(rq->x,"Content-length"));
                        rq->phase = PHASE_BODY;
                    }
                else
                    rq->phase = PHASE_DONE;
            }
            cur = cbuffer + i + 1;
        }
    }

    if(rq->phase == PHASE_BODY)
    {
        if(i)
        { /* reset if crud on beginning */
            cbuffer = cbuffer + i;
            bufsz = bufsz - i;
        }
        if(bufsz >= rq->clen)
        {
            xmlnode_insert_cdata(rq->x,cbuffer,rq->clen);
            rq->phase = PHASE_DONE;
        }else{
            rq->clen -= bufsz;
        }
    }

    if(rq->phase == PHASE_DONE)
    {
        svc_http_parse(rq);
        svc_http_breakdance(rq);
    }

    if(rq->phase == PHASE_HEAD && i) /* the head[er] was cut off */
    {
        free(rq->buffer);
        rq->buffer = strdup(cbuffer + i);
    }
}


/* new connection, setup our handlers */
void svc_http_connect(thread t, int sock, struct sockaddr_in sa)
{
    thread new;
    hrq rq;
    tstream ts;

    log_debug(ZONE,"svc_http connection established (fd: %d, ip: %s, port: %d)\n",
                sock, inet_ntoa(sa.sin_addr), ntohs(sa.sin_port));
    new = tstream_new(sock, inet_ntoa(sa.sin_addr), svc_http_read, NULL, NULL);
    ts = (tstream)(new->data);
    rq = pmalloc(new->p, sizeof(_hrq));
    memset(rq,0,sizeof(_hrq));
    rq->t = new;
    rq->p = t->p;
    rq->ip = ts->host;
    rq->re = xmlnode_new_tag_pool(t->p,"re");
    ts->data = (void *)rq;
}

/* startup routine for jserver */
void svc_http()
{
    log_debug(ZONE,"svc_http startup");
    tlisten_new(5280, NULL, svc_http_connect, NULL);

    svc_http_public_init();
}

void svc_http_qs2xml(xmlnode x, char *qs)
{
    char *key, *val, *nul;

    val = nul = qs + strlen(qs);
    key = qs;
    for(;qs[0] != '\0'; ++qs)
    {
        if(qs[0] == '+')
            qs[0] = ' ';
        if(qs[0] == '=')
        {
            qs[0] = '\0';
            val = qs + 1;
        }
        if(qs[0] == '&' || qs[1] == '\0')
        {
            if(qs[0] == '&')
                qs[0] = '\0';
            svc_http_strdecode(key);
            svc_http_strdecode(val);
            xmlnode_put_attrib(x,key,val);
            key = qs + 1;
            val = nul;
        }
    }
}

int _svc_http_dehex(char c)
{
    if(c >= '0' && c <= '9')
        return c - '0';
    if(c >= 'a' && c <= 'f')
        return c - 'a' + 10;
    if(c >= 'A' && c <= 'F')
        return c - 'A' + 10;
    return 0;
}

/* decode the uri hex junk */
void svc_http_strdecode(char *from)
{
    char *to;

    to = from;
    for(;*from != '\0'; ++to, ++from)
    {
        if(from[0] == '%' && strchr("0123456789abcdefABCDEF", from[1]) != NULL && strchr("0123456789abcdefABCDEF", from[2]) != NULL)
        {
            *to = _svc_http_dehex(from[1]) * 16 + _svc_http_dehex(from[2]);
            from += 2;
        }else{
            *to = *from;
        }
    }
    *to = '\0';
}

void svc_http_error(hrq rq, int err)
{
    rq->re_code = err;
    svc_http_reply(rq);
}

void svc_http_redirect(hrq rq, char *url)
{
    if(url == NULL) return;

    rq->re_code = 302;
    xmlnode_put_attrib(rq->re,"Location",url);
    svc_http_reply(rq);
}

void svc_http_reply_mime(hrq rq, int type)
{
    switch(type)
    {
        case MIME_HTML:
            xmlnode_put_attrib(rq->re,"Content-type","text/html");
            break;
        case MIME_JS:
            xmlnode_put_attrib(rq->re,"Content-type","application/x-javascript");
            break;
        case MIME_XML:
            xmlnode_put_attrib(rq->re,"Content-type","text/xml");
            break;
        case MIME_GIF:
            xmlnode_put_attrib(rq->re,"Content-type","image/gif");
            break;
        case MIME_TXT:
            xmlnode_put_attrib(rq->re,"Content-type","text/plain");
            break;
    }
}

void svc_http_reply_append(hrq rq, ...)
{
    va_list ap;
    char *arg = NULL;
    if(rq == NULL)
        return;

    va_start(ap, rq);
    /* loop till we hit our end flag, the first arg */
    while(1)
    {
        arg = va_arg(ap,char *);
        if((int)arg == (int)rq)
            break;
        else if(arg != NULL)
            xmlnode_insert_cdata(rq->re,arg,strlen(arg));
    }
    va_end(ap);
}

void svc_http_reply(hrq rq)
{
    spool s;
    char *body;
    xmlnode cur;
    char len[10];

    s = spool_new(rq->p);
    spool_add(s,"HTTP/1.0 ");
    switch(rq->re_code)
    {
        case 302:
            spool_add(s,"302 Temporary Redirect\r\n");
            svc_http_reply_mime(rq,MIME_HTML);
            svc_http_reply_append(rq,"Redirecting",rq);
            break;
        case 401:
            spool_add(s,"401 Unauthorized\r\n");
            break;
        case 404:
            spool_add(s,"404 Not Found\r\n");
            svc_http_reply_mime(rq,MIME_HTML);
            svc_http_reply_append(rq,"<H1>Not Found</H1>",rq);
            break;
        default:
            spool_add(s,"200 OK\r\n");
    }

    spool_add(s,"Server: jabber/svc_http\r\n");
    spool_add(s,"Connection: close\r\n");

    for(cur = xmlnode_get_firstattrib(rq->re); cur != NULL; cur = xmlnode_get_nextsibling(cur))
        spooler(s,xmlnode_get_name(cur),": ",xmlnode_get_data(cur),"\r\n",s);

    body = xmlnode_get_data(rq->re);
    if(body != NULL)
    {
        sprintf(len,"%d",strlen(body));
        spooler(s,"Content-length: ",len,"\r\n\r\n",s);
        spool_add(s,body);
    }else{
        spool_add(s,"Content-length: 0\r\n\r\n");
    }

    body = spool_print(s);
    log_debug(ZONE,"HTTP response [%s]",body);
    tstream_write(rq->t, body, -1);
    tstream_close(rq->t);
}

