#include "svc_irc.h"

/* this is a big hack for debugging/testing, someone PLEASE turn it something useful :) */

/*
 * 
 * /nick username (only before logging in)
 * /msg register password
 * /msg login password [resource]
 * 
 * /msg roster [show|add|del] [user@server] [nick]
 * 
 * /me [on|off] [status]
 * 
 * shows all roster items as nicks in channel, use just the username if no name attrib on item
 * 
 * use ppdb to track which are online, show as opers
 * type=chat are PRIVMSG to user
 * 
 */

/* send an outgoing line to the connection */
void svc_irc_line(ircdata d, char *from, char *cmd, char *data, int argc, ...)
{
    static char buff[512], args[512];
    char *str, *nl;
    int strsz, i, argsz, flag;
    va_list ap;

    if(from == NULL)
        from = d->server;

    va_start(ap, argc);
    argsz = 0;
    args[argsz] = '\0';
    for(i = 0; i < argc; i++)
    {
        str = va_arg(ap,char *);
        if(str == NULL) continue;
        if(strlen(str) + argsz >= 500) continue;
        strcat(args,str);
        argsz += strlen(str);
        strcat(args," ");
        argsz++;
    }
    va_end(ap);

    if(data == NULL)
    {
        strsz = snprintf(buff,512,":%s %s %s\r\n",from,cmd,args);
        log_debug("IRC-OUT1",buff);
        tstream_write(d->t,buff,strsz);
        return;
    }

    argsz = strlen(data);
    flag = 0;
    for(i = 1; i < argsz; i++)
    {
        if(data[i] == '\n')
        {
            if(data[i-1] == '\r')
                data[i-1] = '\0';
            data[i] = '\0';
            flag = 1;
            strsz = snprintf(buff,512,":%s %s %s:%s\r\n",from,cmd,args,data);
        }else if(i == 200){
            strsz = snprintf(buff,512,":%s %s %s:%.200s\r\n",from,cmd,args,data);
            flag = 1;
        }

        if(flag)
        {
            log_debug("IRC-OUT2",buff);
            tstream_write(d->t,buff,strsz);
            flag = 0;
            data = data + i + 1;
            i = -1; /* get incremented by for */
            argsz = strlen(data);
        }
    }

    if(strlen(data) > 0)
    {
        strsz = snprintf(buff,512,":%s %s %s:%s\r\n",from,cmd,args,data);
        log_debug("IRC-OUT3",buff);
        tstream_write(d->t,buff,strsz);
    }
}

/* when the connection dies, clean up the mess */
void svc_irc_cleanup(void *arg)
{
    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);
    ircdata d = (ircdata)(ts->data);
 
    log_debug(ZONE,"svc_irc_cleanup");
    if(d->s != NULL)
    {
        d->s->arg = NULL;
        js_session_end(d->s,"Disconnected");
    }

    /* unlink from the pinger */
    d->prev->next = d->next;
    d->next->prev = d->prev;
}

ircdata svc_irc_pinger;

/* read an incoming line from the connection */
void svc_irc_read(thread t, void *buffer, int bufsz)
{
    char *cbuffer, *cur, *to, *args, *data;
    tstream ts = (tstream)(t->data);
    ircdata d = (ircdata)(ts->data);
    int i;

    if(d->initflag)
    {
        pth_cleanup_push(svc_irc_cleanup, (void *)t);
        d->initflag = 0;

        /* insert into pinger ring */
        d->next = svc_irc_pinger->next;
        svc_irc_pinger->next->prev = d;
        d->prev = svc_irc_pinger;
        svc_irc_pinger->next = d;
    }

    if(bufsz <= 0) return;

    if(d->buffer != NULL)
    { /* old data yet, combine and use that */
        cur = malloc(strlen(d->buffer) + bufsz + 1);
        *cur = '\0';
        strcat(cur,d->buffer);
        strcat(cur,buffer);
        free(d->buffer);
        d->buffer = buffer = cur;
        bufsz = strlen(buffer);
    }

    log_debug(ZONE,"phase %d data [%s]",d->phase, buffer);

    cur = cbuffer = (char *)buffer;
    for(i = 0; i < bufsz; i++)
    {
        if(cbuffer[i] == '\r' || cbuffer[i] == '\n')
        {
            cbuffer[i] = '\0';
            if(strlen(cur) > 0)
            {
                to = strstr(cur," ");
                if(to != NULL)
                {
                    *to = '\0';
                    ++to;
                }
                if(to != NULL && (args = strstr(to," ")) != NULL)
                {
                    *args = '\0';
                    ++args;
                }
                if(args != NULL && (data = strstr(args,":")) != NULL)
                {
                    *data = '\0';
                    ++data;
                }else if(to != NULL && (data = strstr(to,":")) != NULL){
                    *data = '\0';
                    ++data;
                }
                svc_irc_out(d, cur, to, args, data);
            }
            cur = cbuffer + i + 1;
        }
    }

    if(strlen(cur) > 0) /* the read was cut off */
    {
        free(d->buffer);
        d->buffer = strdup(cur);
    }
}

/* new connection, setup our handlers */
void svc_irc_connect(thread t, int sock, struct sockaddr_in sa)
{
    thread new;
    ircdata newd;
    tstream ts;

    log_debug(ZONE,"svc_irc connection established (fd: %d, ip: %s, port: %d)\n",
                sock, inet_ntoa(sa.sin_addr), ntohs(sa.sin_port));
    new = tstream_new(sock, inet_ntoa(sa.sin_addr), svc_irc_read, NULL, NULL);
    ts = (tstream)(new->data);
    newd = pmalloc(new->p, sizeof(_ircdata));
    memset(newd,0,sizeof(_ircdata));
    newd->initflag = 1;
    newd->t = new;
    newd->server = xmlnode_get_data(js_config("names/default"));

    ts->data = (void *)newd;
}

void *svc_irc_main(void *arg)
{
    ircdata cur;
    svc_irc_pinger = malloc(sizeof(_ircdata)); /* forever */
    svc_irc_pinger->next = svc_irc_pinger->prev = svc_irc_pinger; /* init ring */

    while(1)
    {
        pth_sleep(PINGER);
        for(cur = svc_irc_pinger->next;cur != svc_irc_pinger;cur = cur->next)
            tstream_write(cur->t,"PING :doh!\r\n",12);
    }
}

/* startup routine for jserver */
void svc_irc()
{
    pth_attr_t attr;
    log_debug(ZONE,"svc_irc startup");
    tlisten_new(6667, NULL, svc_irc_connect, NULL);
 
    attr = pth_attr_new();
    pth_attr_set(attr, PTH_ATTR_JOINABLE, FALSE);
    pth_attr_set(attr, PTH_ATTR_STACK_SIZE, 16*1024);
    pth_spawn(attr, svc_irc_main, NULL); /* pings connections */
    pth_attr_destroy(attr);
}


/* utility for checking in roster */
xmlnode svc_irc_item(ircdata d, jid id)
{
    char *query, *res, *nick;
    xmlnode item;

    if(id == NULL) return NULL;

    /* hmm, try w/o the resource first (IMPORTANT) */
    if(id->resource != NULL)
    {
        res = id->resource;
        jid_set(id,NULL,JID_RESOURCE);
        item = svc_irc_item(d, id);
        jid_set(id,res,JID_RESOURCE);
        if(item != NULL)
            return item;
    }

    item = jid_nodescan(id,d->roster);

    for(nick = xmlnode_get_attrib(item,"name"); nick != NULL && nick[0] != '\0'; nick++)
        if(nick[0] == ' ') nick[0] = '_';

    return item;
}

void _svc_irc_nicknice(char *nick, jid id)
{
    char *bad, *full;

    if(id->user != NULL)
    {
        strncat(nick,id->user,25);
        return;
    }

    strncat(nick,jid_full(id),40);
    bad = strpbrk(nick,"@ ?\r\n\t:!");
    if(bad != NULL)
        *bad = '\0';
}


/* utility for making an irc-nick from a jid: nick!user@server/res */
char *svc_irc_nick(ircdata d, jid id)
{
    static char ret[256];
    char *full, *res, nick[50];
    xmlnode item;

    nick[0] = ret[0] = '\0';
    full = jid_full(id);
    item = svc_irc_item(d, id);

    /* first half, the nick */
    if(item != NULL)
    {
        /* reset the full id to the one in the roster item */
        full = xmlnode_get_attrib(item,"jid");
        if(xmlnode_get_attrib(item,"name") == NULL)
        {
            _svc_irc_nicknice(nick,id);
            while(svc_irc_id(d,nick) != NULL)
                strcat(nick,"_");
            xmlnode_put_attrib(item,"name",nick);
        }else{
            strncat(nick,xmlnode_get_attrib(item,"name"),50);
        }
   }else{
        _svc_irc_nicknice(nick,id);
        while(svc_irc_id(d,nick) != NULL)
            strcat(nick,"_");

        /* save the nick association */
        item = xmlnode_insert_tag(d->roster,"item");
        xmlnode_put_attrib(item,"jid",full);
        xmlnode_put_attrib(item,"name",nick);
    }
    strcat(ret,nick);

    /* second half */
    strcat(ret,"!");
    if(id->user == NULL) /* hack to make sure irc validity */
        strcat(ret,"nouser@");
    strncat(ret,full,200);

    /* chop off any resource */
    res = strstr(strstr(ret,"!"),"/");
    if(res != NULL)
        *res = '\0';

    return ret;
}

/* utility for fetching the jabber id from just the nick */
char *svc_irc_id(ircdata d, char *nick)
{
    xmlnode item;

    if(j_strcmp(nick,d->nick) == 0)
        return jid_full(d->id);

    for(item = xmlnode_get_firstchild(d->roster);item != NULL;item = xmlnode_get_nextsibling(item))
    {
        if(j_strcmp(nick,xmlnode_get_attrib(item,"name")) == 0)
            return xmlnode_get_attrib(item,"jid");
    }
}

